# How to become a contributor and submit your own code

## Contributor License Agreements

We'd love to accept your patches! Before we can take them, we
have to jump a couple of legal hurdles.

Please fill out either the individual or corporate Contributor License Agreement
(CLA).

  * If you are an individual writing original source code and you're sure you
    own the intellectual property, then you'll need to sign an
    [individual CLA](https://developers.google.com/open-source/cla/individual).
  * If you work for a company that wants to allow you to contribute your work,
    then you'll need to sign a
    [corporate CLA](https://developers.google.com/open-source/cla/corporate).

Follow either of the two links above to access the appropriate CLA and
instructions for how to sign and return it. Once we receive it, we'll be able to
accept your pull requests.

## Contributing A Patch

1. Submit an issue describing your proposed change to the repo in question.
1. The repo owner will respond to your issue promptly.
1. If your proposed change is accepted, and you haven't already done so, sign a
   Contributor License Agreement (see details above).
1. Fork the desired repo, develop and test your code changes.
1. Ensure that your code adheres to the existing style in the sample to which
   you are contributing.
1. Ensure that your code has an appropriate set of unit tests which all pass.
1. Submit a pull request.

### More Information on Forks and Pull Requests

If you are just getting started with `git` and [GitHub](https://github.com),
we think the following links may help you understand the usual workflow used
in this project:

* [Forking Projects](https://guides.github.com/activities/forking/)
* [Understanding the GitHub Flow](https://guides.github.com/introduction/flow/)
* [Syncing a Fork](https://help.github.com/articles/syncing-a-fork/)
  * A more succinct guide to [Keeping a Fork up to date](https://gist.github.com/CristinaSolana/1885435)

If your pull request has a conflict with the destination branch we request
that you [rebase](https://help.github.com/articles/about-git-rebase/) your
branch against the destination branch. Reviewing PRs that include both existing
and new changes is a source of error that we would rather eliminate.

## Style

This repository follow the [Google C++ Style Guide](
https://google.github.io/styleguide/cppguide.html).
Please make sure your contributions adhere to the style guide.

### Formatting

The code in this project is formatted with `clang-format(1)`, and our CI builds
will check that the code matches the format generated by this tool before
accepting a pull request. Please configure your editor or IDE to use the Google
style for indentation and other whitespace. If you need to reformat one or more
files, you can simply run `clang-format` manually:

```console
$ clang-format -i <file>....
```

Reformatting all the files in a specific directory should be safe too, for 
example:

```console
$ find bigtable -o -name '*.h' -o -name '*.cc' -print0 \
    | xargs -0 clang-format -i
```

If you need to reformat one of the files to match the Google style.  Please be
advised that `clang-format` has been known to generate slightly different
formatting in different versions, we use version 4.0, use the same version if
you run into problems.

If you have prepared a Docker image for `ubuntu:17.10` (see below), you can
verify and fix the format of your code using:

```console
# Run from google-cloud-cpp:
$ TRAVIS_OS_NAME=linux DISTRO=ubuntu DISTRO_VERSION=17.10 CXX=clang++ CC=clang CHECK_STYLE=yes ./ci/build-linux.sh
```

### Updating CMakeLists.txt and/or BUILD files

If you need to change the list of files associated with a library, please change
the existing `CMakeLists.txt` files, even if you use Bazel as your preferred
build system.  Changing the CMake files automatically update the corresponding
`.bzl` files, but changing the `.bzl` files will not. More details in the short
[design doc](doc/working-with-bazel-and-cmake.md).

## Advanced Compilation and Testing

Please see the [README](README.md) for the basic instructions on how to compile
the code.  In this section we will describe some advanced options.

### Changing the Compiler

If your workstation has multiple compilers (or multiple versions of a compiler)
installed, you can change the compiler using:

```console
# Run this in your build directory, typically google-cloud-cpp/.build
$ CXX=clang++ CC=clang cmake ..

# Then compile and test normally:
$ make -j 4
$ make -j 4 test
```

### Changing the Build Type

By default, the system is compiled with optimizations on; if you want to compile
a debug version, use:

```console
# Run this in your build directory, typically google-cloud-cpp/.build
$ CXX=clang++ CC=clang cmake -DCMAKE_BUILD_TYPE=Debug ..

# Then compile and test normally:
$ make -j 4
$ make -j 4 test
```

This project supports `Debug`, `Release`, and `Coverage` builds.

### Using Docker to Compile and Test

This project uses Docker in its CI builds to test against multiple Linux
distributions, using the native compiler and library versions included in those
distributions.
From time to time, you may need to run the same build on your workstation.
To do so, [install Docker](https://docs.docker.com/engine/installation/)
on your workstation, then create the build Docker image for the environment you
intend to use, for example:

```console
# Run from the google-cloud-cpp directory.
$ TRAVIS_OS_NAME=linux DISTRO=ubuntu DISTRO_VERSION=17.10 ./ci/install-linux.sh
```

Once you create the image for a given combination of `DISTRO` and
`DISTRO_VERSION`, you can compile the code multiple times, for example:

```console
# Also run from google-cloud-cpp:
$ TRAVIS_OS_NAME=linux DISTRO=ubuntu DISTRO_VERSION=17.10 CXX=clang++ CC=clang BUILD_TYPE=Debug ./ci/build-linux.sh
```

You can set any of the following environment variables to control the build.
Please consult the build matrix in your [`.travis.yml`](.travis.yml) file to see
which combinations are tested regularly.

 * `CXX`: the C++ compiler, you can use a full path if needed.
 * `CC`: the C compiler, you can use a full path if needed. Typically you
   set both variables to select the compiler. For example:
   * `CXX=clang++ CC=clang` to use Clang.
   * `CXX=g++ CC=gcc` to use GCC.
 * `DISTRO`: the Linux distribution, use `ubuntu`, `fedora`, or `centos`.
 * `DISTRO_VERSION`: the version of the distribution, e.g. `16.04`.
 * `CHECK_STYLE`: if set to `yes`, the build fails if the code is different
   than the output from `clang-format(1)`.  Note that this reformats your files,
   that can be useful to keep the formatting clean.
 * `GENERATE_DOCS`: if set to `yes` the build will generate the documentation
   using [Doxygen](https://www.doxygen.org).  In addition, if `GH_TOKEN` is set
   it will try to update the `gh-pages` branch on your fork with the new
   documentation.
 * `BUILD_TYPE`: if set, override the `CMAKE_BUILD_TYPE` flag when configuring
   the build.
 * `SCAN_BUILD`: if set to `yes`, use Clang static analyzer (aka `scan-build`)
   to compile the code.  Remember to also set the compiler to Clang as described
   above.  Builds with this configuration can be substantially slower.
 * `CMAKE_FLAGS`: if set, these additional cmake flags are used to configure
   the build.  The more interesting flags include:
   * `-DSANITIZE_ADDRESS=yes`: use the
     [AddressSanitizer](https://clang.llvm.org/docs/AddressSanitizer.html),
     if available, on this build.
   * `-DSANITIZE_UNDEFINED=yes`: use the
     [UndefinedBehaviorSanitizer](https://clang.llvm.org/docs/UndefinedBehaviorSanitizer.html),
     if available, on this build.
   * `-DSANITIZE_THREAD=yes`: use the
     [ThreadSanitizer](https://clang.llvm.org/docs/ThreadSanitizer.html),
     if available, on this build.
   * `-DSANITIZE_MEMORY=yes`: use the
     [MemorySanitizer](https://clang.llvm.org/docs/MemorySanitizer.html),
     if available, on this build.
     This is rarely used because it requires a C++ library compiled with the
     same options to avoid false positives.
   * All the sanitizers require Clang, remember to select the compiler as
     described above.
 * `-DGOOGLE_CLOUD_CPP_CLANG_TIDY=yes`: configure the build to run `clang-tidy`
   for the Google Cloud C++ Libraries.  Check the
   [configuration file](.clang-tidy) for details on what `clang-tidy` options
   we use.
